/***************************************************************************
                          mypoly3d.h  -  description
                             -------------------
    begin                : Thu Jun 28 2001
    copyright            : (C) 2001 by Rdiger Knrig
    email                : ruediger@knoerig.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef MYPOLY3D_H
#define MYPOLY3D_H


#include <string.h>
#include <math.h>

/**
	* provides data structures for modelling
	* 3d objects bases on triangles
	* the 3D-system is a right-oriented (X and Y in screen orientaion, Z points in the screen)
 	*	@author Rdiger Knrig
 	*/

class MyPoly3D {
public: 
	MyPoly3D() {}
	~MyPoly3D() {}
	/** structure representing a color in red,green,blue */
	class color3f
	{
	public:
		float r,g,b;
		color3f(float r=0.0,float g=0.0,float b=0.0) { this->r=r; this->g=g; this->b=b; }
		~color3f() {}
		color3f operator*(float s) { return color3f(r*s,g*s,b*s); }
		color3f operator*(const color3f &c) { return color3f(r*c.r,g*c.g,b*c.b); }
		color3f operator+(const color3f &c) { return color3f(r+c.r,g+c.g,b+c.b); }
		void operator+=(const color3f &c) { r+=c.r; g+=c.g; b+=c.b; }
	};
	/**
		* structure representing a polygon
		* @param p1 index of first vertex
		* @param p2 index of second vertex
		* @param p3 index of third vertex
		* @param mindex index of used material
		*/
	struct poly {int p1,p2,p3; int mindex; };
	/** class representing a point in homogeneous 3D-coordinates */
	class point4f;
	/** class representing the transformation matrix */
	class tmatrix;
};

static const float entity[4][4] = {{1.0,0.0,0.0,0.0},
											 			  {0.0,1.0,0.0,0.0},
											  		      {0.0,0.0,1.0,0.0},
											 			  {0.0,0.0,0.0,1.0}};
											 			
											 			
/**
	* This class provides a 4x4 transformation matrix for both view and object transformations
	* @author Rdiger Knrig
	*/
class MyPoly3D::tmatrix
{
	friend MyPoly3D::point4f operator*(MyPoly3D::point4f v, MyPoly3D::tmatrix m);
	
	public:
		/** supported view transformations */
		enum  viewtransform_t {orthoXY=0,orthoYZ=1,orthoXZ=2,persp1=3};
		/** default constructor, generates an identity matrix */
		tmatrix() { reset(); }
		/** copy constructor */
		tmatrix(const MyPoly3D::tmatrix &src) { memcpy(matrix,src.matrix,16*sizeof(float)); }
		/** default destruktor */
		virtual ~tmatrix() {}
		/** resets matrix to entity matrix */
		void reset() { memcpy(matrix,entity,16*sizeof(float)); }
		/**
			* set view transformation type
			* @param type View transformation type
			* @param bw Optional focal length of camera lens (perspective transformation)
			*/
		void setTransform(viewtransform_t type, float bw = 1.0);
		
		/**
			* add a translation to the matrix
			* @param x X component of translation vector
			* @param y Y component of translation vector
			* @param z Z component of translation vector
			*/
		void translate(float x, float y, float z);
		
		/**
			* add a scaling operation to the matrix
			* @param x X scaling factor
			* @param y Y scaling factor
			* @param z Z scaling factor
			*/
		void scale(float,float y,float z);
		
		/**
			* Rotation around x axis
			* @param degrees Rotation angle in degrees
			*/
		void rotateAroundX(float degrees);
		
		/**
			* Rotation around y axis
			* @param degrees Rotation angle in degrees
			*/
		void rotateAroundY(float degrees);
		
		/**
			* Rotation around z axis
			* @param degrees Rotation angle in degrees
			*/
		void rotateAroundZ(float degrees);
		
		/** multiplication: matrix = matrix * matrix */
		MyPoly3D::tmatrix operator*(MyPoly3D::tmatrix b);
		
	protected:
		float matrix[4][4];
};

/** multiplication: vector = vector * matrix */
MyPoly3D::point4f operator*(MyPoly3D::point4f v, MyPoly3D::tmatrix m);

/** scalar product of two vectors */
float  operator*(MyPoly3D::point4f v1, MyPoly3D::point4f v2);

/** if you see point4f as a vector in homogenous 3D coordinates, this function calulates its magnitude */
double abs(MyPoly3D::point4f);



/**
	* this class represents a point in homogenous 3D coordinates
	* @author Rdiger Knrig
	*/
class MyPoly3D::point4f
{
	friend class MyPoly3D::tmatrix;
	friend MyPoly3D::point4f operator*(MyPoly3D::point4f v, MyPoly3D::tmatrix m);
	friend double abs(MyPoly3D::point4f);
	friend MyPoly3D::point4f crossp(const MyPoly3D::point4f &v1,const MyPoly3D::point4f &v2);
	friend float  operator*(MyPoly3D::point4f v1, MyPoly3D::point4f v2);
	
	public:
		/** default constructor, results in a 0/0/0/0 - point */
		point4f() { memset(data,0,4*sizeof(float)); data[3]=1.0;}
		/** copy constructor */
		point4f(const MyPoly3D::point4f &src) { memcpy(data,src.data,4*sizeof(float)); }
		/** constructor from 4 float values (w is optional ) */
		point4f(float x, float y, float z, float w=1.0) { data[0]=x; data[1]=y; data[2]=z; data[3]=w; }
		/** default destructor */
		virtual ~point4f() {}
		/** method for access to x component */
		float &x() {return data[0];}
		/** method for access to y component */
		float &y() {return data[1];}
		/** method for access to z component */
		float &z() {return data[2];}
		/** method for access to w component */
		float &w() {return data[3];}
		/** substraction this - b = result */
		point4f operator-(point4f b) {return point4f(data[0]/data[3]-b.data[0]/b.data[3],data[1]/data[3]-b.data[1]/b.data[3],data[2]/data[3]-b.data[2]/b.data[3],1.0); }
		/** addition this + b = result */
		point4f operator+(point4f b) {return point4f(data[0]/data[3]+b.data[0]/b.data[3],data[1]/data[3]+b.data[1]/b.data[3],data[2]/data[3]+b.data[2]/b.data[3],1.0); }		
		/** normalize (x/=w,y/=w,z/=w,w=1) */
		void normalize() { data[0] /= data[3]; data[1] /= data[3]; data[2] /= data[3]; data[3] = 1.0; }
		/** normalize & scale to a magnitude of 1  */
		void makenormv() { float temp =  (data[3]*data[3])/sqrt(data[0]*data[0] + data[1]*data[1] + data[2]*data[2]);  data[0] *= temp; data[1] *= temp; data[2] *= temp; data[3]=1.0; }
		/** returns the magnitude when point4f is interpreted as a 3D vector */
		double magnitude() const { return sqrt(data[0]*data[0] + data[1]*data[1] + data[2]*data[2])/data[3]; }
		/** addition of another vector */
		void operator+=(MyPoly3D::point4f v) { data[0]=data[0]/data[3]+v.data[0]/v.data[3];data[1]=data[1]/data[3]+v.data[1]/v.data[3];data[2]=data[2]/data[3]+v.data[2]/v.data[3];data[3]=1.0; }
		/** quick addition of points width w=1 to the current point */
		void quickAdd(const point4f &b) { data[0]+=b.data[0];  data[1]+=b.data[1];  data[2]+=b.data[2]; }
		/** quick substraction of points width w=1 to the current point */
		void quickSub(const point4f &b) { data[0]-=b.data[0];  data[1]-=b.data[1];  data[2]-=b.data[2]; }
		/**  quick multiplication for points width w=1 and a scalar */
		void quickMul(float c) { data[0]*=c;  data[1]*=c;  data[2]*=c; }
		/**  quick division for points width w=1 and a scalar */
		void quickDiv(float c) { float temp = 1.0/c; data[0]*=temp;  data[1]*=temp;  data[2]*=temp; }
	protected:
		float data[4];	
};

/**
	* calculates the cross-product of two vectors
	* @param v1 first vector
	* @param v2 second vector
	* @return crossproduct vector v1xv2
	*/
MyPoly3D::point4f crossp(const MyPoly3D::point4f &v1,const MyPoly3D::point4f &v2);


#endif
